import numpy as np
from typing import Optional, Tuple

__all__ = ["DatasetMeanStdCalculator"]


class DatasetMeanStdCalculator:
    __slots__ = ["global_mean", "global_var", "n_items", "num_channels", "global_max", "global_min", "dtype"]

    """
    Class to calculate running mean and std of the dataset. It helps when whole dataset does not fit entirely in RAM.
    """

    def __init__(self, num_channels: int = 3, dtype=np.float64):
        """
        Create a new instance of DatasetMeanStdCalculator

        Args:
            num_channels: Number of channels in the image. Default value is 3
        """
        super().__init__()
        self.num_channels = num_channels
        self.global_mean = None
        self.global_var = None
        self.global_max = None
        self.global_min = None
        self.n_items = 0
        self.dtype = dtype
        self.reset()

    def reset(self):
        self.global_mean = np.zeros(self.num_channels, dtype=self.dtype)
        self.global_var = np.zeros(self.num_channels, dtype=self.dtype)
        self.global_max = np.ones_like(self.global_mean) * float("-inf")
        self.global_min = np.ones_like(self.global_mean) * float("+inf")
        self.n_items = 0

    def accumulate(self, image: np.ndarray, mask: Optional[np.ndarray] = None) -> None:
        """
        Compute mean and std of a single image and integrates it into global statistics
        Args:
            image: Input image (Must be HWC, with number of channels C equal to self.num_channels)
            mask: Optional mask to include only certain parts of image from statistics computation.
            Only non-zero elements will be included,
        """
        if len(image.shape) == 2:
            image = np.expand_dims(image, axis=-1)

        if self.num_channels != image.shape[2]:
            raise RuntimeError(f"Number of channels in image must be {self.num_channels}, got {image.shape[2]}.")
        image = image.reshape((-1, self.num_channels))

        if mask is not None:
            mask = mask.reshape((mask.shape[0] * mask.shape[1]))
            image = image[mask, :]

            # In case the whole image is masked out, we exclude it entirely
            if len(image) == 0:
                return

        mean = np.mean(image, axis=0)
        std = np.std(image, axis=0)

        self.global_mean += np.squeeze(mean)
        self.global_var += np.squeeze(std) ** 2
        self.global_max = np.maximum(self.global_max, np.max(image, axis=0))
        self.global_min = np.minimum(self.global_min, np.min(image, axis=0))
        self.n_items += 1

    def compute(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute dataset-level mean & std

        Returns:
            Tuple of global [mean, std] per channel
        """
        return self.global_mean / self.n_items, np.sqrt(self.global_var / self.n_items)



from __future__ import absolute_import

import math

import torch.nn as nn

from ..activations import get_activation_block


def conv_bn(inp, oup, stride, activation):
    return nn.Sequential(nn.Conv2d(inp, oup, 3, stride, 1, bias=False), nn.BatchNorm2d(oup), activation())


def conv_1x1_bn(inp, oup, activation):
    return nn.Sequential(nn.Conv2d(inp, oup, 1, 1, 0, bias=False), nn.BatchNorm2d(oup), activation())


class InvertedResidual(nn.Module):
    def __init__(self, inp, oup, stride, expand_ratio, activation):
        super(InvertedResidual, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        hidden_dim = round(inp * expand_ratio)
        self.use_res_connect = self.stride == 1 and inp == oup

        if expand_ratio == 1:
            self.conv = nn.Sequential(
                # dw
                nn.Conv2d(hidden_dim, hidden_dim, 3, stride, 1, groups=hidden_dim, bias=False),
                nn.BatchNorm2d(hidden_dim),
                activation(),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup),
            )
        else:
            self.conv = nn.Sequential(
                # pw
                nn.Conv2d(inp, hidden_dim, 1, 1, 0, bias=False),
                nn.BatchNorm2d(hidden_dim),
                activation(),
                # dw
                nn.Conv2d(hidden_dim, hidden_dim, 3, stride, 1, groups=hidden_dim, bias=False),
                nn.BatchNorm2d(hidden_dim),
                activation(),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup),
            )

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


class MobileNetV2(nn.Module):
    def __init__(self, n_class=1000, input_size=224, width_mult=1.0, activation="relu6"):
        super(MobileNetV2, self).__init__()

        activation_block = get_activation_block(activation)

        block = InvertedResidual
        input_channel = 32
        last_channel = 1280
        interverted_residual_setting = [
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 2],
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1],
        ]

        # building first layer
        assert input_size % 32 == 0
        input_channel = int(input_channel * width_mult)
        self.last_channel = int(last_channel * width_mult) if width_mult > 1.0 else last_channel
        self.layer0 = conv_bn(3, input_channel, 2, activation_block)

        # building inverted residual blocks
        for layer_index, (t, c, n, s) in enumerate(interverted_residual_setting):
            output_channel = int(c * width_mult)

            blocks = []
            for i in range(n):
                if i == 0:
                    blocks.append(block(input_channel, output_channel, s, expand_ratio=t, activation=activation_block))
                else:
                    blocks.append(block(input_channel, output_channel, 1, expand_ratio=t, activation=activation_block))

                input_channel = output_channel

            self.add_module(f"layer{layer_index + 1}", nn.Sequential(*blocks))

        # building last several layers
        self.final_layer = conv_1x1_bn(input_channel, self.last_channel, activation=activation_block)

        # building classifier
        self.classifier = nn.Sequential(nn.Dropout(0.2), nn.Linear(self.last_channel, n_class))

        self._initialize_weights()

    def forward(self, x):
        x = self.layer0(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.layer5(x)
        x = self.layer6(x)
        x = self.layer7(x)
        x = self.final_layer(x)

        x = x.mean(3).mean(2)
        x = self.classifier(x)
        return x

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                n = m.weight.size(1)
                m.weight.data.normal_(0, 0.01)
                m.bias.data.zero_()



from __future__ import absolute_import

from .balanced_bce import *
from .bitempered_loss import *
from .dice import *
from .focal import *
from .focal_cosine import *
from .functional import *
from .jaccard import *
from .lovasz import *
from .soft_bce import *
from .soft_ce import *
from .soft_f1 import *
from .wing_loss import *
from .logcosh import *
from .quality_focal_loss import *



import dataclasses
from abc import abstractmethod
from typing import Protocol, Tuple, List, Union, Mapping, Callable

import numpy as np
import torch.jit

__all__ = [
    "FeatureMapsSpecification",
    "HasOutputFeaturesSpecification",
    "AbstractDecoder",
    "AbstractHead",
    "AbstractEncoder",
]

from torch import nn, Tensor

from pytorch_toolbelt.utils import pytorch_toolbelt_deprecated


@dataclasses.dataclass
class FeatureMapsSpecification:
    channels: Tuple[int, ...]
    strides: Tuple[int, ...]

    def __init__(self, channels: Union[Tuple[int, ...], List[int]], strides: Union[Tuple[int, ...], List[int]]):
        if len(channels) != len(strides):
            raise RuntimeError(
                f"Length of feature_map_channels({len(channels)} must"
                f" be equal to length of feature_map_strides({len(strides)})"
            )

        self.channels = tuple(channels)
        self.strides = tuple(strides)

    def get_index_of_largest_feature_map(self) -> int:
        """
        Returns index of largest (spatially) feature map (with smallest stride)
        :return: 0-based index of largest feature map
        """
        return int(np.argmin(self.strides))

    def get_dummy_input(self, device=None, image_size=(640, 512)) -> List[Tensor]:
        """
        Returns dummy input for this feature map specification
        :return: Dummy input tensor
        """
        feature_maps = []
        rows, cols = image_size
        for c, s in zip(self.channels, self.strides):
            feature_maps.append(torch.randn((1, c, rows // s, cols // s), device=device))
        return feature_maps

    def __len__(self) -> int:
        return len(self.channels)


class HasInputFeaturesSpecification(Protocol):
    """
    A protocol for modules that have output features
    """

    @torch.jit.unused
    def get_input_spec(self) -> FeatureMapsSpecification:
        ...


class HasOutputFeaturesSpecification(Protocol):
    """
    A protocol for modules that have output features
    """

    @torch.jit.unused
    def get_output_spec(self) -> FeatureMapsSpecification:
        ...


class AbstractEncoder(nn.Module, HasOutputFeaturesSpecification):
    pass


class AbstractDecoder(nn.Module, HasInputFeaturesSpecification, HasOutputFeaturesSpecification):
    __constants__ = ["input_spec"]

    def __init__(self, input_spec: FeatureMapsSpecification):
        if input_spec is None:
            raise ValueError("input_spec must be specified")
        super().__init__()
        self.input_spec = input_spec

    def forward(self, feature_maps: List[Tensor]) -> List[Tensor]:  # skipcq: PYL-W0221
        raise NotImplementedError

    @torch.jit.unused
    def set_trainable(self, trainable):
        for param in self.parameters():
            param.requires_grad = bool(trainable)

    @torch.jit.unused
    def get_input_spec(self):
        return self.input_spec


class AbstractHead(AbstractDecoder):
    def __init__(self, input_spec: FeatureMapsSpecification):
        super().__init__(input_spec)

    @abstractmethod
    def forward(
        self, feature_maps: List[Tensor], output_size: Union[Tuple[int, int], torch.Size, None] = None
    ) -> Union[Tensor, Tuple[Tensor, ...], List[Tensor], Mapping[str, Tensor]]:
        ...

    @torch.jit.unused
    def apply_to_final_layer(self, func: Callable[[nn.Module], None]):
        """
        Apply function to the final layer of the head.

        Typically, you can use this method to apply custom initialization
        to the final layer of the head.

        :param func: Function to apply to the final prediction layer. If head contains
        supervision layers, function can be applied to them as well.

        """
        raise NotImplementedError("This method is not implemented in class " + self.__class__.__name__)



import torch
from pytorch_toolbelt.losses.functional import log_cosh_loss
from torch import nn

__all__ = ["LogCoshLoss"]


class LogCoshLoss(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        return log_cosh_loss(y_pred, y_true)



from typing import List

import torch
import torch.nn.functional as F
from pytorch_toolbelt.utils.torch_utils import to_tensor
from torch import Tensor
from torch.nn.modules.loss import _Loss

from .functional import soft_jaccard_score

__all__ = ["JaccardLoss", "BINARY_MODE", "MULTICLASS_MODE", "MULTILABEL_MODE"]

BINARY_MODE = "binary"
MULTICLASS_MODE = "multiclass"
MULTILABEL_MODE = "multilabel"


class JaccardLoss(_Loss):
    """
    Implementation of Jaccard loss for image segmentation task.
    It supports binary, multi-class and multi-label cases.
    """

    def __init__(self, mode: str, classes: List[int] = None, log_loss=False, from_logits=True, smooth=0, eps=1e-7):
        """

        :param mode: Metric mode {'binary', 'multiclass', 'multilabel'}
        :param classes: Optional list of classes that contribute in loss computation;
        By default, all channels are included.
        :param log_loss: If True, loss computed as `-log(jaccard)`; otherwise `1 - jaccard`
        :param from_logits: If True assumes input is raw logits
        :param smooth:
        :param eps: Small epsilon for numerical stability
        """
        assert mode in {BINARY_MODE, MULTILABEL_MODE, MULTICLASS_MODE}
        super(JaccardLoss, self).__init__()
        self.mode = mode
        if classes is not None:
            assert mode != BINARY_MODE, "Masking classes is not supported with mode=binary"
            classes = to_tensor(classes, dtype=torch.long)

        self.classes = classes
        self.from_logits = from_logits
        self.smooth = smooth
        self.eps = eps
        self.log_loss = log_loss

    def forward(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        """

        :param y_pred: NxCxHxW
        :param y_true: NxHxW
        :return: scalar
        """
        assert y_true.size(0) == y_pred.size(0)

        if self.from_logits:
            # Apply activations to get [0..1] class probabilities
            # Using Log-Exp as this gives more numerically stable result and does not cause vanishing gradient on
            # extreme values 0 and 1
            if self.mode == MULTICLASS_MODE:
                y_pred = y_pred.log_softmax(dim=1).exp()
            else:
                y_pred = F.logsigmoid(y_pred).exp()

        bs = y_true.size(0)
        num_classes = y_pred.size(1)
        dims = (0, 2)

        if self.mode == BINARY_MODE:
            y_true = y_true.view(bs, 1, -1)
            y_pred = y_pred.view(bs, 1, -1)

        if self.mode == MULTICLASS_MODE:
            y_true = y_true.view(bs, -1)
            y_pred = y_pred.view(bs, num_classes, -1)

            y_true = F.one_hot(y_true, num_classes)  # N,H*W -> N,H*W, C
            y_true = y_true.permute(0, 2, 1)  # H, C, H*W

        if self.mode == MULTILABEL_MODE:
            y_true = y_true.view(bs, num_classes, -1)
            y_pred = y_pred.view(bs, num_classes, -1)

        scores = soft_jaccard_score(y_pred, y_true.type(y_pred.dtype), smooth=self.smooth, eps=self.eps, dims=dims)

        if self.log_loss:
            loss = -torch.log(scores.clamp_min(self.eps))
        else:
            loss = 1.0 - scores

        # IoU loss is defined for non-empty classes
        # So we zero contribution of channel that does not have true pixels
        # NOTE: A better workaround would be to use loss term `mean(y_pred)`
        # for this case, however it will be a modified jaccard loss

        mask = y_true.sum(dims) > 0
        loss *= mask.float()

        if self.classes is not None:
            loss = loss[self.classes]

        return loss.mean()



import collections
from typing import List, Dict

import torch.nn
from torch import nn
from torch.optim import SGD

from pytorch_toolbelt.optimization.functional import build_optimizer_param_groups
from pytorch_toolbelt.utils import count_parameters


def count_parameters_in_param_groups(pg: List[Dict[str, List[torch.nn.Parameter]]]) -> Dict[str, int]:
    kv_iter = enumerate(pg)
    dict_iter = [(str(key), sum([p.numel() for p in value["params"]])) for key, value in kv_iter]
    return dict(dict_iter)


def test_build_optimizer_param_groups():
    def conv_bn_relu(in_channels, out_channels):
        return nn.Sequential(
            collections.OrderedDict(
                [
                    ("conv", nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1)),
                    ("bn", nn.BatchNorm2d(out_channels)),
                    ("relu", nn.ReLU(inplace=True)),
                ]
            )
        )

    model = nn.Sequential(
        collections.OrderedDict(
            [
                ("encoder", nn.Sequential(conv_bn_relu(3, 32), conv_bn_relu(32, 64))),
                ("neck", nn.Sequential(conv_bn_relu(64, 64), conv_bn_relu(64, 64).requires_grad_(False))),
                ("decoder", nn.Sequential(conv_bn_relu(64, 32), conv_bn_relu(32, 1))),
            ]
        )
    )

    pg, defaults = build_optimizer_param_groups(
        model, learning_rate=1e-4, weight_decay=0, apply_weight_decay_on_bias=False, apply_weight_decay_on_norm=False
    )

    total_params = count_parameters(model)
    total_params_in_pg = count_parameters_in_param_groups(pg)
    assert SGD(pg, **defaults) is not None

    print(total_params)
    print(total_params_in_pg)
    assert len(pg) == 3
    assert sum(total_params_in_pg.values()) == total_params["trainable"]

    pg, defaults = build_optimizer_param_groups(
        model,
        learning_rate={"encoder": 1e-3, "neck": 1e-4, "decoder": 1e-5, "_default_": 0},
        weight_decay={
            "encoder.0": 1e-5,
            "encoder.1": 5e-5,
            "_default_": 1e-6,
        },
        apply_weight_decay_on_bias=False,
        apply_weight_decay_on_norm=False,
    )

    print(total_params)
    print(total_params_in_pg)
    assert len(pg) == 10
    assert sum(total_params_in_pg.values()) == total_params["trainable"]



import ast
import io
import re
import os
import sys

from setuptools import find_packages, setup
from distutils.version import LooseVersion


def is_docker() -> bool:
    """
    Check whether setup is running in Docker environment.
    """
    # Note: You have to set the environment variable AM_I_IN_A_DOCKER_CONTAINER manually
    # in your Dockerfile .
    if os.environ.get("AM_I_IN_A_DOCKER_CONTAINER", False):
        return True

    path = "/proc/self/cgroup"
    if not os.path.isfile(path):
        return False

    with open(path) as f:
        for line in f:
            if re.match("\\d+:[\\w=]+:/docker(-[ce]e)?/\\w+", line):
                return True

    return False


def is_kaggle() -> bool:
    """
    Check whether setup is running in Kaggle environment.
    This is not 100% bulletproff solution to detect whether we are in Kaggle Notebooks,
    but it should be enough unless Kaggle change their environment variables.
    """
    return (
        ("KAGGLE_CONTAINER_NAME" in os.environ)
        or ("KAGGLE_URL_BASE" in os.environ)
        or ("KAGGLE_DOCKER_IMAGE" in os.environ)
    )


def is_colab() -> bool:
    """
    Check whether setup is running in Google Colab.
    This is not 100% bulletproff solution to detect whether we are in Colab,
    but it should be enough unless Google change their environment variables.
    """
    return (
        ("COLAB_GPU" in os.environ)
        or ("GCE_METADATA_TIMEOUT" in os.environ)
        or ("GCS_READ_CACHE_BLOCK_SIZE_MB" in os.environ)
    )


def get_opencv_requirement():
    """
    Return the OpenCV requirement string.
    Since opencv library is distributed in several independent packages,
    we first check whether any form of opencv is already installed. If not,
    we choose between opencv-python vs opencv-python-headless version based
    on the environment.
    For headless environment (Docker, Colab & Kaggle notebooks), we install
    opencv-python-headless; otherwise - default to opencv-python.
    """
    try:
        import cv2

        return []
    except ImportError:
        default_requirement = "opencv-python>=4.1"
        headless_requirement = "opencv-python-headless>=4.1"

        if is_docker() or is_kaggle() or is_colab():
            return [headless_requirement]
        else:
            return [default_requirement]


# Package meta-data.
NAME = "pytorch_toolbelt"
DESCRIPTION = "PyTorch extensions for fast R&D prototyping and Kaggle farming"
URL = "https://github.com/BloodAxe/pytorch-toolbelt"
EMAIL = "ekhvedchenya@gmail.com"
AUTHOR = "Eugene Khvedchenya"
REQUIRES_PYTHON = ">=3.8.0"

DEPENDENCIES = [
    # We rely on particular activation functions that were added in 1.8.1
    "torch>=1.11.0" if LooseVersion(sys.version) >= LooseVersion("3.7") else "torch>=1.10.0",
    # We use some pretrained models from torchvision
    "torchvision",
    # Library uses scipy for linear_sum_assignment for match_bboxes.
    # 1.4.0 is the first release where `maximize` argument gets introduced to this function
    "scipy>=1.4.0",
] + get_opencv_requirement()

EXCLUDE_FROM_PACKAGES = ["contrib", "docs", "tests", "examples"]
CURDIR = os.path.abspath(os.path.dirname(__file__))


def get_version():
    main_file = os.path.join(CURDIR, "pytorch_toolbelt", "__init__.py")
    _version_re = re.compile(r"__version__\s+=\s+(?P<version>.*)")
    with open(main_file, "r", encoding="utf8") as f:
        match = _version_re.search(f.read())
        version = match.group("version") if match is not None else '"unknown"'
    return str(ast.literal_eval(version))


def load_readme():
    readme_path = os.path.join(CURDIR, "README.md")
    with io.open(readme_path, encoding="utf-8") as f:
        return "\n" + f.read()


def get_test_requirements():
    requirements = ["pytest", "black==23.3.0", "timm==0.6.7", "matplotlib"]
    return requirements


setup(
    name=NAME,
    version=get_version(),
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description=load_readme(),
    long_description_content_type="text/markdown",
    url=URL,
    packages=find_packages(exclude=EXCLUDE_FROM_PACKAGES),
    install_requires=DEPENDENCIES,
    python_requires=REQUIRES_PYTHON,
    extras_require={"tests": get_test_requirements()},
    include_package_data=True,
    keywords=[
        "PyTorch",
        "Kaggle",
        "Deep Learning",
        "Machine Learning",
        "ResNet",
        "VGG",
        "ResNext",
        "Unet",
        "Focal",
        "FPN",
        "EfficientNet",
        "Test-Time Augmentation",
        "Model Ensembling",
        "Model Surgery",
    ],
    scripts=[],
    license="License :: OSI Approved :: MIT License",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        # "Private :: Do Not Upload"
    ],
)



import collections

import torch
from torch import nn, Tensor
from typing import List, Union, Iterable, Optional, Dict, Tuple

__all__ = ["ApplySoftmaxTo", "ApplySigmoidTo", "Ensembler", "PickModelOutput", "SelectByIndex", "average_checkpoints"]

from pytorch_toolbelt.inference.tta import _deaugment_averaging


class ApplySoftmaxTo(nn.Module):
    output_keys: Tuple
    temperature: float
    dim: int

    def __init__(
        self,
        model: nn.Module,
        output_key: Union[str, int, Iterable[str]] = "logits",
        dim: int = 1,
        temperature: float = 1,
    ):
        """
        Apply softmax activation on given output(s) of the model
        :param model: Model to wrap
        :param output_key: string, index or list of strings, indicating to what outputs softmax activation should be applied.
        :param dim: Tensor dimension for softmax activation
        :param temperature: Temperature scaling coefficient. Values > 1 will make logits sharper.
        """
        super().__init__()
        # By converting to set, we prevent double-activation by passing output_key=["logits", "logits"]
        output_key = (output_key,) if isinstance(output_key, (str, int)) else tuple(set(output_key))
        self.output_keys = output_key
        self.model = model
        self.dim = dim
        self.temperature = temperature

    def forward(self, *input, **kwargs):
        output = self.model(*input, **kwargs)
        for key in self.output_keys:
            output[key] = output[key].mul(self.temperature).softmax(dim=self.dim)
        return output


class ApplySigmoidTo(nn.Module):
    output_keys: Tuple
    temperature: float

    def __init__(self, model: nn.Module, output_key: Union[str, int, Iterable[str]] = "logits", temperature=1):
        """
        Apply sigmoid activation on given output(s) of the model
        :param model: Model to wrap
        :param output_key: string index, or list of strings, indicating to what outputs sigmoid activation should be applied.
        :param temperature: Temperature scaling coefficient. Values > 1 will make logits sharper.
        """
        super().__init__()
        # By converting to set, we prevent double-activation by passing output_key=["logits", "logits"]
        output_key = (output_key,) if isinstance(output_key, (str, int)) else tuple(set(output_key))
        self.output_keys = output_key
        self.model = model
        self.temperature = temperature

    def forward(self, *input, **kwargs):  # skipcq: PYL-W0221
        output = self.model(*input, **kwargs)
        for key in self.output_keys:
            output[key] = output[key].mul(self.temperature).sigmoid_()
        return output


class Ensembler(nn.Module):
    __slots__ = ["outputs", "reduction", "return_some_outputs"]

    """
    Compute sum (or average) of outputs of several models.
    """

    def __init__(self, models: List[nn.Module], reduction: str = "mean", outputs: Optional[Iterable[str]] = None):
        """

        :param models:
        :param reduction: Reduction key ('mean', 'sum', 'gmean', 'hmean' or None)
        :param outputs: Name of model outputs to average and return from Ensembler.
            If None, all outputs from the first model will be used.
        """
        super().__init__()
        self.return_some_outputs = outputs is not None
        self.outputs = tuple(outputs) if outputs else tuple()
        self.models = nn.ModuleList(models)
        self.reduction = reduction

    def forward(self, *input, **kwargs):  # skipcq: PYL-W0221
        outputs = [model(*input, **kwargs) for model in self.models]
        output_is_dict = isinstance(outputs[0], dict)
        output_is_list = isinstance(outputs[0], (list, tuple))  # noqa

        if self.return_some_outputs:
            keys = self.outputs
        elif output_is_dict:
            keys = outputs[0].keys()
        elif output_is_list:
            keys = list(range(len(outputs[0])))
        elif torch.is_tensor(outputs[0]):
            keys = None
        else:
            raise RuntimeError()

        if keys is None:
            predictions = torch.stack(outputs)
            predictions = _deaugment_averaging(predictions, self.reduction)
            averaged_output = predictions
        else:
            averaged_output = {} if output_is_dict else []
            for key in keys:
                predictions = [output[key] for output in outputs]
                predictions = torch.stack(predictions)
                predictions = _deaugment_averaging(predictions, self.reduction)
                if output_is_dict:
                    averaged_output[key] = predictions
                else:
                    averaged_output.append(predictions)

        return averaged_output


class PickModelOutput(nn.Module):
    """
    Wraps a model that returns dict or list and returns only a specific element.

    Usage example:
        >>> model = MyAwesomeSegmentationModel() # Returns dict {"OUTPUT_MASK": Tensor, ...}
        >>> net  = nn.Sequential(PickModelOutput(model, "OUTPUT_MASK")), nn.Sigmoid())
    """

    __slots__ = ["target_key"]

    def __init__(self, model: nn.Module, key: Union[str, int]):
        super().__init__()
        self.model = model
        self.target_key = key

    def forward(self, *input, **kwargs) -> Tensor:
        output = self.model(*input, **kwargs)
        return output[self.target_key]


class SelectByIndex(nn.Module):
    """
    Select a single Tensor from the dict or list of output tensors.

    Usage example:
        >>> model = MyAwesomeSegmentationModel() # Returns dict {"OUTPUT_MASK": Tensor, ...}
        >>> net  = nn.Sequential(model, SelectByIndex("OUTPUT_MASK"), nn.Sigmoid())
    """

    __slots__ = ["target_key"]

    def __init__(self, key: Union[str, int]):
        super().__init__()
        self.target_key = key

    def forward(self, outputs: Dict[str, Tensor]) -> Tensor:
        return outputs[self.target_key]


def average_checkpoints(inputs: List[str]) -> collections.OrderedDict:
    """Loads checkpoints from inputs and returns a model with averaged weights. Original implementation taken from:
    https://github.com/pytorch/fairseq/blob/a48f235636557b8d3bc4922a6fa90f3a0fa57955/scripts/average_checkpoints.py#L16
    Args:
      inputs (List[str]): An iterable of string paths of checkpoints to load from.
    Returns:
      A dict of string keys mapping to various values. The 'model' key
      from the returned dict should correspond to an OrderedDict mapping
      string parameter names to torch Tensors.
    """
    params_dict = collections.OrderedDict()
    params_keys = None
    new_state = None
    num_models = len(inputs)
    for fpath in inputs:
        with open(fpath, "rb") as f:
            state = torch.load(
                f,
                map_location="cpu",
            )
        # Copies over the settings from the first checkpoint
        if new_state is None:
            new_state = state
        model_params = state["model_state_dict"]
        model_params_keys = list(model_params.keys())
        if params_keys is None:
            params_keys = model_params_keys
        elif params_keys != model_params_keys:
            raise KeyError(
                "For checkpoint {}, expected list of params: {}, "
                "but found: {}".format(f, params_keys, model_params_keys)
            )
        for k in params_keys:
            p = model_params[k]
            if isinstance(p, torch.HalfTensor):
                p = p.float()
            if k not in params_dict:
                params_dict[k] = p.clone()
                # NOTE: clone() is needed in case of p is a shared parameter
            else:
                params_dict[k] += p
    averaged_params = collections.OrderedDict()
    for k, v in params_dict.items():
        averaged_params[k] = v
        if averaged_params[k].is_floating_point():
            averaged_params[k].div_(num_models)
        else:
            averaged_params[k] //= num_models
    new_state["model_state_dict"] = averaged_params
    return new_state



from torch import nn

__all__ = ["DropPath"]


def drop_path(x, drop_prob: float = 0.0, training: bool = False, scale_by_keep: bool = True):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.

    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
    if keep_prob > 0.0 and scale_by_keep:
        random_tensor.div_(keep_prob)
    return x * random_tensor


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob: float = 0.0, scale_by_keep: bool = True):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob
        self.scale_by_keep = scale_by_keep

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training, self.scale_by_keep)

    def extra_repr(self):
        return f"drop_prob={round(self.drop_prob,3):0.3f}"



from typing import List, Tuple, Union, Type, Callable

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from pytorch_toolbelt.modules.activations import (
    ACT_RELU,
    instantiate_activation_block,
)
from pytorch_toolbelt.modules.simple import conv1x1
from pytorch_toolbelt.modules.dsconv import DepthwiseSeparableConv2dBlock
from pytorch_toolbelt.modules.interfaces import AbstractDecoder, FeatureMapsSpecification
from pytorch_toolbelt.modules.normalization import NORM_BATCH, instantiate_normalization_block

__all__ = ["BiFPNDecoder", "BiFPNBlock", "BiFPNConvBlock"]


class BiFPNConvBlock(nn.Module):
    """
    Convolution block with Batch Normalization and ReLU activation.

    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        stride=1,
        padding=1,
        activation: str = ACT_RELU,
        dilation=1,
        normalization: str = NORM_BATCH,
    ):
        super(BiFPNConvBlock, self).__init__()

        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, stride=stride, padding=padding, bias=False)
        self.bn = instantiate_normalization_block(normalization, out_channels)
        self.act = instantiate_activation_block(activation, inplace=True)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        x = self.bn(x)
        return self.act(x)


class BiFPNBlock(nn.Module):
    """
    Bi-Directional Feature Pyramid Network
    """

    __constants__ = ["epsilon"]

    def __init__(
        self,
        feature_size: int,
        num_feature_maps: int,
        epsilon=0.0001,
        activation=ACT_RELU,
        normalization="batch",
        block: Union[Type[BiFPNConvBlock], Type[DepthwiseSeparableConv2dBlock]] = BiFPNConvBlock,
    ):
        super(BiFPNBlock, self).__init__()

        self.epsilon = epsilon
        num_blocks = num_feature_maps - 1

        self.top_down_blocks = nn.ModuleList(
            [
                block(feature_size, feature_size, activation=activation, normalization=normalization)
                for _ in range(num_blocks)
            ]
        )
        self.bottom_up_blocks = nn.ModuleList(
            [
                block(feature_size, feature_size, activation=activation, normalization=normalization)
                for _ in range(num_blocks)
            ]
        )

        self.register_parameter("w1", nn.Parameter(torch.Tensor(2, num_blocks), requires_grad=True))
        self.register_parameter("w2", nn.Parameter(torch.Tensor(3, num_blocks), requires_grad=True))

        torch.nn.init.constant_(self.w1, 1)
        torch.nn.init.constant_(self.w2, 1)

    def forward(self, inputs: List[Tensor]) -> List[Tensor]:
        transition_features = self.top_down_pathway(inputs)
        return self.bottom_up_pathway(transition_features, inputs)

    def bottom_up_pathway(self, transition_features, inputs):
        # Calculate Bottom-Up Pathway
        # p3_out = p3_td
        # p4_out = self.p4_out(
        #     w2[0, 0] * p4_x + w2[1, 0] * p4_td + w2[2, 0] * F.interpolate(p3_out, size=p4_x.size()[2:])
        # )
        # p5_out = self.p5_out(
        #     w2[0, 1] * p5_x + w2[1, 1] * p5_td + w2[2, 1] * F.interpolate(p4_out, size=p5_x.size()[2:])
        # )
        # p6_out = self.p6_out(
        #     w2[0, 2] * p6_x + w2[1, 2] * p6_td + w2[2, 2] * F.interpolate(p5_out, size=p6_x.size()[2:])
        # )
        # p7_out = self.p7_out(
        #     w2[0, 3] * p7_x + w2[1, 3] * p7_td + w2[2, 3] * F.interpolate(p6_out, size=p7_x.size()[2:])
        # )

        w2 = torch.nn.functional.relu(self.w2)
        w2 = w2 / (torch.sum(w2, dim=0) + self.epsilon)

        outputs = [transition_features[-1]]
        transition_reversed = transition_features[:-1][::-1]
        for i, block in enumerate(self.bottom_up_blocks):
            x = inputs[i + 1]
            td = transition_reversed[i]
            y = block(x * w2[0, i] + td * w2[1, i] + F.interpolate(outputs[-1], size=x.size()[2:]) * w2[2, i])
            outputs.append(y)

        return outputs

    def top_down_pathway(self, inputs: List[Tensor]) -> List[Tensor]:
        # p6_td = self.p6_td(w1[0, 0] * p6_x + w1[1, 0] * F.interpolate(p7_td, size=p6_x.size()[2:]))
        # p5_td = self.p5_td(w1[0, 1] * p5_x + w1[1, 1] * F.interpolate(p6_td, size=p5_x.size()[2:]))
        # p4_td = self.p4_td(w1[0, 2] * p4_x + w1[1, 2] * F.interpolate(p5_td, size=p4_x.size()[2:]))
        # p3_td = self.p3_td(w1[0, 3] * p3_x + w1[1, 3] * F.interpolate(p4_td, size=p3_x.size()[2:]))

        w1 = torch.nn.functional.relu(self.w1)
        w1 = w1 / (torch.sum(w1, dim=0) + self.epsilon)

        features = [inputs[-1]]
        inputs_reversed = inputs[:-1][::-1]

        for i, block in enumerate(self.top_down_blocks):
            x = inputs_reversed[i]
            y = block(w1[0, i] * x + w1[1, i] * F.interpolate(features[-1], size=x.size()[2:]))
            features.append(y)

        return features


class BiFPNDecoder(AbstractDecoder):
    """
    BiFPN decoder. Note this class does not compute additional feature maps (p6/p7) and expects them to be present in input.
    Because of this it supports arbitrary number of input feature maps.

    Reference: https://arxiv.org/abs/1911.09070
    """

    __constants__ = ["output_spec"]

    def __init__(
        self,
        input_spec: FeatureMapsSpecification,
        out_channels: int,
        num_layers: int,
        activation: str = ACT_RELU,
        normalization: str = NORM_BATCH,
        block: Union[
            Type[BiFPNConvBlock], Type[DepthwiseSeparableConv2dBlock], Callable[[int, int], nn.Module]
        ] = BiFPNConvBlock,
        projection_block: Callable[[int, int], nn.Module] = conv1x1,
    ):
        super().__init__(input_spec)

        self.projections = nn.ModuleList(
            [projection_block(in_channels, out_channels) for in_channels in input_spec.channels]
        )

        bifpns = []
        for _ in range(num_layers):
            bifpns.append(
                BiFPNBlock(
                    out_channels,
                    num_feature_maps=len(input_spec),
                    activation=activation,
                    normalization=normalization,
                    block=block,
                )
            )
        self.bifpn = nn.Sequential(*bifpns)

        self.output_spec = FeatureMapsSpecification(
            channels=tuple([out_channels] * len(input_spec)), strides=input_spec.strides
        )

    def forward(self, feature_maps: List[Tensor]) -> List[Tensor]:
        # Calculate the input column of BiFPN
        features = [p(c) for p, c in zip(self.projections, feature_maps)]
        return self.bifpn(features)

    @torch.jit.unused
    def get_output_spec(self) -> FeatureMapsSpecification:
        return self.output_spec



import dataclasses
import numbers
from typing import Optional, Iterator, Dict, Union, List, Tuple, Mapping

from torch import nn
from pytorch_toolbelt.utils.distributed import get_rank, get_world_size, is_dist_avail_and_initialized
from pytorch_toolbelt.utils.torch_utils import get_non_wrapped_model

__all__ = ["scale_learning_rate_for_ddp", "get_optimizable_parameters", "freeze_model", "build_optimizer_param_groups"]


def scale_learning_rate_for_ddp(
    learning_rate: Union[numbers.Rational, Dict[str, numbers.Rational]]
) -> Union[float, Dict[str, float]]:
    """
    Scale learning rate with respect to world size for Distributed Data Parallel training.
    Efficient learning rate is WORLD_SIZE * learning rate from config
    For non-distributed runs it is No-Op and returns learning_rate argument as is.
    """
    if not is_dist_avail_and_initialized():
        return learning_rate

    scale = float(get_world_size())
    if isinstance(learning_rate, Mapping):
        return dict((k, float(v * scale)) for k, v in learning_rate.items())
    elif isinstance(learning_rate, (float, numbers.Rational)):
        return scale * learning_rate
    raise ValueError(
        f"Got unsupported type {type(learning_rate)} for learning rate. Must be either a mapping or a single scalar."
    )


@dataclasses.dataclass
class ParametersGroup:
    lr: Union[float, None]
    weight_decay: Union[float, None]
    params: List
    name: str

    def __len__(self):
        return len(self.params)

    def asdict(self):
        d = {"params": self.params, "name": self.name}
        if self.lr is not None:
            d["lr"] = self.lr
        if self.weight_decay is not None:
            d["weight_decay"] = self.weight_decay
        return d


def recursive_getattr(obj, attr):
    """Get object's attribute. May use dot notation.

    >>> class C(object): pass
    >>> a = C()
    >>> a.b = C()
    >>> a.b.c = 4
    >>> recursive_getattr(a, 'b.c')
    4
    """
    if "." not in attr:
        return getattr(obj, attr)
    else:
        L = attr.split(".")
        return recursive_getattr(getattr(obj, L[0]), ".".join(L[1:]))


def build_optimizer_param_groups(
    model: nn.Module,
    learning_rate: Union[float, Mapping[str, float]],
    weight_decay: Union[float, Mapping[str, float]],
    apply_weight_decay_on_bias: bool = True,
    apply_weight_decay_on_norm: bool = True,
) -> Tuple[List[Dict], Dict]:
    """

    Args:
        model: A model whose parameters will be used to fill corresponding parameter groups.
        learning_rate: A single number of dictionary of layer-wise learning rate parameters.
        weight_decay: A single number of dictionary of layer-wise weight decay parameters.
        apply_weight_decay_on_bias: If True, weight decay is applied to bias on Linear & Conv layers (default).
            This parameter is False, it overrules the matching layer-wise weight-decay parameter.
        apply_weight_decay_on_norm: If True, weight decay is applied normalization layers (default).
            This parameter is False, it overrules the matching layer-wise weight-decay parameter.

    Returns:

    """
    model = get_non_wrapped_model(model)

    if isinstance(learning_rate, Mapping) and "_default_" not in learning_rate:
        raise RuntimeError(
            "When using layerwise learning rate, a key _default_ must be present to indicate default LR"
        )

    if isinstance(weight_decay, Mapping) and "_default_" not in weight_decay:
        raise RuntimeError("When using layerwise weight decay, a key _default_ must be present to indicate default LR")

    if isinstance(learning_rate, numbers.Number):
        learning_rate = {"_default_": float(learning_rate)}
    if isinstance(weight_decay, numbers.Number):
        weight_decay = {"_default_": float(weight_decay)}

    default_learning_rate = float(learning_rate["_default_"])
    default_weight_decay = float(weight_decay["_default_"])

    learning_rate = list(learning_rate.items())
    weight_decay = list(weight_decay.items())

    parameter_groups = {}

    norm_layers = (
        nn.BatchNorm1d,
        nn.BatchNorm2d,
        nn.BatchNorm3d,
        nn.GroupNorm,
        nn.LayerNorm,
        nn.InstanceNorm1d,
        nn.InstanceNorm3d,
        nn.InstanceNorm2d,
        nn.SyncBatchNorm,
    )

    layers_with_bias = (
        nn.Linear,
        nn.Conv1d,
        nn.Conv2d,
        nn.Conv3d,
        nn.ConvTranspose1d,
        nn.ConvTranspose2d,
        nn.ConvTranspose3d,
    )

    def get_param_group(parameter_name: str, module: nn.Module):
        matching_lr_index = "default"
        matching_lr_value = default_learning_rate
        for lr_index, (prefix, lr) in enumerate(learning_rate):
            if parameter_name.startswith(prefix):
                matching_lr_index = prefix
                matching_lr_value = lr
                break

        matching_wd_index = "default"
        matching_wd_value = default_weight_decay
        for wd_index, (prefix, wd) in enumerate(weight_decay):
            if parameter_name.startswith(prefix):
                matching_wd_index = prefix
                matching_wd_value = wd
                break

        if not apply_weight_decay_on_norm and isinstance(module, norm_layers):
            matching_wd_value = 0
            matching_wd_index = "no_wd_on_norm"

        if (
            not apply_weight_decay_on_bias
            and isinstance(module, layers_with_bias)
            and parameter_name.endswith(".bias")
        ):
            matching_wd_value = 0
            matching_wd_index = "no_wd_on_bias"

        if matching_lr_index == matching_wd_index:
            param_group_name = f"{matching_lr_index}"
        else:
            param_group_name = f"{matching_lr_index}_{matching_wd_index}"
        if param_group_name not in parameter_groups:
            parameter_groups[param_group_name] = ParametersGroup(
                lr=matching_lr_value, weight_decay=matching_wd_value, name=param_group_name, params=[]
            )

        return parameter_groups[param_group_name]

    # All optimizable parameters
    parameters = get_named_optimizable_parameters(model)
    total_optimizable_params = 0

    for parameter_name, parameter in parameters:
        total_optimizable_params += parameter.numel()
        module_name = ".".join(parameter_name.split(".")[:-1]) if "." in parameter_name else parameter_name
        module = recursive_getattr(model, module_name)

        param_group: ParametersGroup = get_param_group(parameter_name, module)
        param_group.params.append(parameter)

    defaults = {"lr": default_learning_rate, "weight_decay": default_weight_decay}
    param_groups = [x.asdict() for x in parameter_groups.values()]

    total_params_count_from_groups = 0
    for pg in param_groups:
        total_params_count_from_groups += sum(x.numel() for x in pg["params"])

    if total_params_count_from_groups != total_optimizable_params:
        raise RuntimeError(
            f"Detected mismatch in total number of optimizable parameters ({total_optimizable_params}) and"
            f"number of parameters across each groups ({total_params_count_from_groups})."
            f"This is likely indicate a bug in build_optimizer_param_groups."
            f"Please report a bug to https://github.com/BloodAxe/pytorch-toolbelt/issues/new?assignees=&labels=&template=bug-report.md"
        )
    return param_groups, defaults


def get_optimizable_parameters(model: nn.Module) -> Iterator[nn.Parameter]:
    """
    Return list of parameters with requires_grad=True from the model.
    This function allows easily get all parameters that should be optimized.
    :param model: An instance of nn.Module.
    :return: Parameters with requires_grad=True.
    """
    return filter(lambda x: x.requires_grad, model.parameters())


def get_named_optimizable_parameters(model: nn.Module, prefix: str = "") -> Iterator[Tuple[str, nn.Parameter]]:
    """
    Return list of parameters with requires_grad=True from the model.
    This function allows easily get all parameters that should be optimized.
    :param model: An instance of nn.Module.
    :return: Parameters with requires_grad=True.
    """
    return filter(lambda x: x[1].requires_grad, model.named_parameters(prefix))


def freeze_model(
    module: nn.Module, freeze_parameters: Optional[bool] = True, freeze_bn: Optional[bool] = True
) -> nn.Module:
    """
    Change 'requires_grad' value for module and it's child modules and
    optionally freeze batchnorm modules.
    :param module: Module to change
    :param freeze_parameters: True to freeze parameters; False - to enable parameters optimization.
        If None - current state is not changed.
    :param freeze_bn: True to freeze batch norm; False - to enable BatchNorm updates.
        If None - current state is not changed.
    :return: None
    """
    bn_types = nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm

    if freeze_parameters is not None:
        for param in module.parameters():
            param.requires_grad = not freeze_parameters

    if freeze_bn is not None:
        if isinstance(module, bn_types):
            module.track_running_stats = not freeze_bn

        for m in module.modules():
            if isinstance(m, bn_types):
                module.track_running_stats = not freeze_bn

    return module



# Adapted from https://github.com/NVlabs/SegFormer
import math
from functools import partial
from typing import Optional

import torch
import torch.nn as nn
from pytorch_toolbelt.modules.interfaces import AbstractEncoder, FeatureMapsSpecification
from pytorch_toolbelt.modules.activations import instantiate_activation_block, ACT_GELU
from pytorch_toolbelt.modules.drop_path import DropPath


class Mlp(nn.Module):
    def __init__(self, in_features, activation, hidden_features=None, out_features=None, drop=0.0):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.dwconv = DWConv(hidden_features)
        self.act = instantiate_activation_block(activation)
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            torch.nn.init.trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        x = self.fc1(x)
        x = self.dwconv(x, H, W)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0.0, proj_drop=0.0, sr_ratio=1):
        super().__init__()
        if dim % num_heads != 0:
            raise ValueError(f"dim {dim} should be divisible by num_heads {num_heads}.")

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5

        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.sr_ratio = sr_ratio
        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, kernel_size=sr_ratio, stride=sr_ratio)
            self.norm = nn.LayerNorm(dim)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            torch.nn.init.trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        B, N, C = x.shape
        q = self.q(x).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)

        if self.sr_ratio > 1:
            x_ = x.permute(0, 2, 1).reshape(B, C, H, W)
            x_ = self.sr(x_).reshape(B, C, -1).permute(0, 2, 1)
            x_ = self.norm(x_)
            kv = self.kv(x_).reshape(B, -1, 2, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        else:
            kv = self.kv(x).reshape(B, -1, 2, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class Block(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio,
        qkv_bias: bool,
        qk_scale: Optional[float],
        drop: float,
        attn_drop: float,
        drop_path: float,
        activation: str,
        sr_ratio: int,
        norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
            sr_ratio=sr_ratio,
        )
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, activation=activation, drop=drop)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            torch.nn.init.trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x, H, W):
        """

        :param x: B, H * W, C
        :param H: Height
        :param W: Width
        :return: B, H * W, C
        """
        x = x + self.drop_path(self.attn(self.norm1(x), H, W))
        x = x + self.drop_path(self.mlp(self.norm2(x), H, W))

        return x


class OverlapPatchEmbed(nn.Module):
    """Image to Patch Embedding"""

    def __init__(self, patch_size=7, stride=4, in_chans=3, embed_dim=768):
        super().__init__()

        self.embed_dim = embed_dim
        self.patch_size = patch_size
        self.proj = nn.Conv2d(
            in_chans,
            embed_dim,
            kernel_size=patch_size,
            stride=stride,
            padding=patch_size // 2,
        )
        self.norm = nn.LayerNorm(embed_dim)
        self.stride = stride
        self.apply(self._init_weights)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.proj = nn.Conv2d(
            input_channels,
            self.embed_dim,
            kernel_size=self.patch_size,
            stride=self.stride,
            padding=self.patch_size // 2,
        )
        return self

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            torch.nn.init.trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def forward(self, x):
        x = self.proj(x)
        _, _, H, W = x.shape
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)

        return x, H, W


class MixVisionTransformer(AbstractEncoder):
    def __init__(
        self,
        in_channels=3,
        embed_dims=[64, 128, 256, 512],
        num_heads=[1, 2, 4, 8],
        mlp_ratios=[4, 4, 4, 4],
        qkv_bias=False,
        qk_scale=None,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.0,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        depths=[3, 4, 6, 3],
        sr_ratios=[8, 4, 2, 1],
        activation=ACT_GELU,
        first_embedd_stride=4,
    ):
        super().__init__()
        self.depths = depths

        # patch_embed
        self.patch_embed1 = OverlapPatchEmbed(
            patch_size=7, stride=first_embedd_stride, in_chans=in_channels, embed_dim=embed_dims[0]
        )
        self.patch_embed2 = OverlapPatchEmbed(patch_size=3, stride=2, in_chans=embed_dims[0], embed_dim=embed_dims[1])
        self.patch_embed3 = OverlapPatchEmbed(patch_size=3, stride=2, in_chans=embed_dims[1], embed_dim=embed_dims[2])
        self.patch_embed4 = OverlapPatchEmbed(patch_size=3, stride=2, in_chans=embed_dims[2], embed_dim=embed_dims[3])
        self.embed_dims = embed_dims

        # transformer encoder
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]  # stochastic depth decay rule
        cur = 0
        self.block1 = nn.ModuleList(
            [
                Block(
                    dim=embed_dims[0],
                    num_heads=num_heads[0],
                    mlp_ratio=mlp_ratios[0],
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    drop=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[cur + i],
                    norm_layer=norm_layer,
                    activation=activation,
                    sr_ratio=sr_ratios[0],
                )
                for i in range(depths[0])
            ]
        )
        self.norm1 = norm_layer(embed_dims[0])

        cur += depths[0]
        self.block2 = nn.ModuleList(
            [
                Block(
                    dim=embed_dims[1],
                    num_heads=num_heads[1],
                    mlp_ratio=mlp_ratios[1],
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    drop=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[cur + i],
                    norm_layer=norm_layer,
                    activation=activation,
                    sr_ratio=sr_ratios[1],
                )
                for i in range(depths[1])
            ]
        )
        self.norm2 = norm_layer(embed_dims[1])

        cur += depths[1]
        self.block3 = nn.ModuleList(
            [
                Block(
                    dim=embed_dims[2],
                    num_heads=num_heads[2],
                    mlp_ratio=mlp_ratios[2],
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    drop=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[cur + i],
                    norm_layer=norm_layer,
                    activation=activation,
                    sr_ratio=sr_ratios[2],
                )
                for i in range(depths[2])
            ]
        )
        self.norm3 = norm_layer(embed_dims[2])

        cur += depths[2]
        self.block4 = nn.ModuleList(
            [
                Block(
                    dim=embed_dims[3],
                    num_heads=num_heads[3],
                    mlp_ratio=mlp_ratios[3],
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    drop=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[cur + i],
                    norm_layer=norm_layer,
                    activation=activation,
                    sr_ratio=sr_ratios[3],
                )
                for i in range(depths[3])
            ]
        )
        self.norm4 = norm_layer(embed_dims[3])

        self.apply(self._init_weights)
        self.output_spec = FeatureMapsSpecification(
            channels=embed_dims,
            strides=(first_embedd_stride, first_embedd_stride * 2, first_embedd_stride * 4, first_embedd_stride * 8),
        )

    @torch.jit.unused
    def change_input_channels(self, input_channels):
        self.patch_embed1.change_input_channels(input_channels)
        return self

    @torch.jit.unused
    def get_output_spec(self):
        return self.output_spec

    @torch.jit.unused
    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            torch.nn.init.trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out //= m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                m.bias.data.zero_()

    def reset_drop_path(self, drop_path_rate):
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(self.depths))]
        cur = 0
        for i in range(self.depths[0]):
            self.block1[i].drop_path.drop_prob = dpr[cur + i]

        cur += self.depths[0]
        for i in range(self.depths[1]):
            self.block2[i].drop_path.drop_prob = dpr[cur + i]

        cur += self.depths[1]
        for i in range(self.depths[2]):
            self.block3[i].drop_path.drop_prob = dpr[cur + i]

        cur += self.depths[2]
        for i in range(self.depths[3]):
            self.block4[i].drop_path.drop_prob = dpr[cur + i]

    def forward_features(self, x):
        B = x.shape[0]
        outs = []

        # stage 1
        x, H, W = self.patch_embed1(x)
        for i, blk in enumerate(self.block1):
            x = blk(x, H, W)
        x = self.norm1(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        # stage 2
        x, H, W = self.patch_embed2(x)
        for i, blk in enumerate(self.block2):
            x = blk(x, H, W)
        x = self.norm2(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        # stage 3
        x, H, W = self.patch_embed3(x)
        for i, blk in enumerate(self.block3):
            x = blk(x, H, W)
        x = self.norm3(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        # stage 4
        x, H, W = self.patch_embed4(x)
        for i, blk in enumerate(self.block4):
            x = blk(x, H, W)
        x = self.norm4(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()
        outs.append(x)

        return outs

    def forward(self, x):
        x = self.forward_features(x)
        return x


class DWConv(nn.Module):
    def __init__(self, dim=768):
        super(DWConv, self).__init__()
        self.dwconv = nn.Conv2d(dim, dim, 3, 1, 1, bias=True, groups=dim)

    def forward(self, x, H, W):
        B, N, C = x.shape
        x = x.transpose(1, 2).view(B, C, H, W)
        x = self.dwconv(x)
        x = x.flatten(2).transpose(1, 2)

        return x


# These weights are ported from official implementation
# https://github.com/NVlabs/SegFormer
MitPretrainedWeights = {
    "mit_b0": "https://github.com/BloodAxe/pytorch-toolbelt/releases/download/0.7.0/pytorch_toolbelt_mit_b0_imagenet.pth",
    "mit_b1": "https://github.com/BloodAxe/pytorch-toolbelt/releases/download/0.7.0/pytorch_toolbelt_mit_b1_imagenet.pth",
    "mit_b2": "https://github.com/BloodAxe/pytorch-toolbelt/releases/download/0.7.0/pytorch_toolbelt_mit_b2_imagenet.pth",
    "mit_b3": "https://github.com/BloodAxe/pytorch-toolbelt/releases/download/0.7.0/pytorch_toolbelt_mit_b3_imagenet.pth",
    "mit_b4": "https://github.com/BloodAxe/pytorch-toolbelt/releases/download/0.7.0/pytorch_toolbelt_mit_b4_imagenet.pth",
    "mit_b5": "https://github.com/BloodAxe/pytorch-toolbelt/releases/download/0.7.0/pytorch_toolbelt_mit_b5_imagenet.pth",
}


class MitB0Encoder(MixVisionTransformer):
    def __init__(
        self, pretrained: bool = True, drop_rate=0.0, drop_path_rate=0.1, activation=ACT_GELU, first_embedd_stride=4
    ):
        super().__init__(
            embed_dims=[32, 64, 160, 256],
            num_heads=[1, 2, 5, 8],
            mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True,
            norm_layer=partial(nn.LayerNorm, eps=1e-6),
            depths=[2, 2, 2, 2],
            sr_ratios=[8, 4, 2, 1],
            drop_rate=drop_rate,
            drop_path_rate=drop_path_rate,
            activation=activation,
            first_embedd_stride=first_embedd_stride,
        )
        if pretrained:
            checkpoint = torch.hub.load_state_dict_from_url(
                MitPretrainedWeights["mit_b0"], map_location="cpu", check_hash=True
            )
            self.load_state_dict(checkpoint["model_state_dict"])


class MitB1Encoder(MixVisionTransformer):
    def __init__(
        self, pretrained: bool = True, drop_rate=0.0, drop_path_rate=0.1, activation=ACT_GELU, first_embedd_stride=4
    ):
        super().__init__(
            embed_dims=[64, 128, 320, 512],
            num_heads=[1, 2, 5, 8],
            mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True,
            norm_layer=partial(nn.LayerNorm, eps=1e-6),
            depths=[2, 2, 2, 2],
            sr_ratios=[8, 4, 2, 1],
            drop_rate=drop_rate,
            drop_path_rate=drop_path_rate,
            activation=activation,
            first_embedd_stride=first_embedd_stride,
        )
        if pretrained:
            checkpoint = torch.hub.load_state_dict_from_url(
                MitPretrainedWeights["mit_b1"], map_location="cpu", check_hash=True
            )
            self.load_state_dict(checkpoint["model_state_dict"])


class MitB2Encoder(MixVisionTransformer):
    def __init__(
        self, pretrained: bool = True, drop_rate=0.0, drop_path_rate=0.1, activation=ACT_GELU, first_embedd_stride=4
    ):
        super().__init__(
            embed_dims=[64, 128, 320, 512],
            num_heads=[1, 2, 5, 8],
            mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True,
            norm_layer=partial(nn.LayerNorm, eps=1e-6),
            depths=[3, 4, 6, 3],
            sr_ratios=[8, 4, 2, 1],
            drop_rate=drop_rate,
            drop_path_rate=drop_path_rate,
            activation=activation,
            first_embedd_stride=first_embedd_stride,
        )
        if pretrained:
            checkpoint = torch.hub.load_state_dict_from_url(
                MitPretrainedWeights["mit_b2"], map_location="cpu", check_hash=True
            )
            self.load_state_dict(checkpoint["model_state_dict"])


class MitB3Encoder(MixVisionTransformer):
    def __init__(
        self, pretrained: bool = True, drop_rate=0.0, drop_path_rate=0.1, activation=ACT_GELU, first_embedd_stride=4
    ):
        super().__init__(
            embed_dims=[64, 128, 320, 512],
            num_heads=[1, 2, 5, 8],
            mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True,
            norm_layer=partial(nn.LayerNorm, eps=1e-6),
            depths=[3, 4, 18, 3],
            sr_ratios=[8, 4, 2, 1],
            drop_rate=drop_rate,
            drop_path_rate=drop_path_rate,
            activation=activation,
            first_embedd_stride=first_embedd_stride,
        )
        if pretrained:
            checkpoint = torch.hub.load_state_dict_from_url(
                MitPretrainedWeights["mit_b3"], map_location="cpu", check_hash=True
            )
            self.load_state_dict(checkpoint["model_state_dict"])


class MitB4Encoder(MixVisionTransformer):
    def __init__(
        self, pretrained: bool = True, drop_rate=0.0, drop_path_rate=0.1, activation=ACT_GELU, first_embedd_stride=4
    ):
        super().__init__(
            embed_dims=[64, 128, 320, 512],
            num_heads=[1, 2, 5, 8],
            mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True,
            norm_layer=partial(nn.LayerNorm, eps=1e-6),
            depths=[3, 8, 27, 3],
            sr_ratios=[8, 4, 2, 1],
            drop_rate=drop_rate,
            drop_path_rate=drop_path_rate,
            activation=activation,
            first_embedd_stride=first_embedd_stride,
        )
        if pretrained:
            checkpoint = torch.hub.load_state_dict_from_url(
                MitPretrainedWeights["mit_b4"], map_location="cpu", check_hash=True
            )
            self.load_state_dict(checkpoint["model_state_dict"])


class MitB5Encoder(MixVisionTransformer):
    def __init__(
        self, pretrained: bool = True, drop_rate=0.0, drop_path_rate=0.1, activation=ACT_GELU, first_embedd_stride=4
    ):
        super().__init__(
            embed_dims=[64, 128, 320, 512],
            num_heads=[1, 2, 5, 8],
            mlp_ratios=[4, 4, 4, 4],
            qkv_bias=True,
            norm_layer=partial(nn.LayerNorm, eps=1e-6),
            depths=[3, 6, 40, 3],
            sr_ratios=[8, 4, 2, 1],
            drop_rate=drop_rate,
            drop_path_rate=drop_path_rate,
            activation=activation,
            first_embedd_stride=first_embedd_stride,
        )

        if pretrained:
            checkpoint = torch.hub.load_state_dict_from_url(
                MitPretrainedWeights["mit_b5"], map_location="cpu", check_hash=True
            )
            self.load_state_dict(checkpoint["model_state_dict"])



from __future__ import absolute_import

from .deeplab_v3 import *
from .deeplab_v3_plus import *
from .fpn import *
from .pyramid_pooling import *
from .unet import *
from .can import *
from .bifpn import *



"""Wrappers for different backbones for models that follows Encoder-Decoder architecture.

Encodes listed here provides easy way to swap backbone of classification/segmentation/detection model.
"""

from .common import *
from .densenet import *
from .hrnet import *
from .inception import *
from .mobilenet import *
from .resnet import *
from .seresnet import *
from .squeezenet import *
from .unet import *
from .wide_resnet import *
from .xresnet import *
from .hourglass import *
from .timm import *
from .swin import *
from .mix_transformer import *



# Orignal source
# https://github.com/Cadene/pretrained-models.pytorch/blob/master/pretrainedmodels/models/inceptionv4.py

from __future__ import print_function, division, absolute_import

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.model_zoo as model_zoo

__all__ = ["InceptionV4", "inceptionv4"]

pretrained_settings = {
    "inceptionv4": {
        "imagenet": {
            "url": "http://data.lip6.fr/cadene/pretrainedmodels/inceptionv4-8e4777a0.pth",
            "input_space": "RGB",
            "input_size": [3, 299, 299],
            "input_range": [0, 1],
            "mean": [0.5, 0.5, 0.5],
            "std": [0.5, 0.5, 0.5],
            "num_classes": 1000,
        },
        "imagenet+background": {
            "url": "http://data.lip6.fr/cadene/pretrainedmodels/inceptionv4-8e4777a0.pth",
            "input_space": "RGB",
            "input_size": [3, 299, 299],
            "input_range": [0, 1],
            "mean": [0.5, 0.5, 0.5],
            "std": [0.5, 0.5, 0.5],
            "num_classes": 1001,
        },
    }
}


class BasicConv2d(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, stride, padding=0):
        super(BasicConv2d, self).__init__()
        self.conv = nn.Conv2d(
            in_planes, out_planes, kernel_size=kernel_size, stride=stride, padding=padding, bias=False
        )  # verify bias false
        self.bn = nn.BatchNorm2d(
            out_planes, eps=0.001, momentum=0.1, affine=True  # value found in tensorflow  # default pytorch value
        )
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class Mixed_3a(nn.Module):
    def __init__(self):
        super(Mixed_3a, self).__init__()
        self.maxpool = nn.MaxPool2d(3, stride=2)
        self.conv = BasicConv2d(64, 96, kernel_size=3, stride=2)

    def forward(self, x):
        x0 = self.maxpool(x)
        x1 = self.conv(x)
        out = torch.cat((x0, x1), 1)
        return out


class Mixed_4a(nn.Module):
    def __init__(self):
        super(Mixed_4a, self).__init__()

        self.branch0 = nn.Sequential(
            BasicConv2d(160, 64, kernel_size=1, stride=1), BasicConv2d(64, 96, kernel_size=3, stride=1)
        )

        self.branch1 = nn.Sequential(
            BasicConv2d(160, 64, kernel_size=1, stride=1),
            BasicConv2d(64, 64, kernel_size=(1, 7), stride=1, padding=(0, 3)),
            BasicConv2d(64, 64, kernel_size=(7, 1), stride=1, padding=(3, 0)),
            BasicConv2d(64, 96, kernel_size=(3, 3), stride=1),
        )

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        out = torch.cat((x0, x1), 1)
        return out


class Mixed_5a(nn.Module):
    def __init__(self):
        super(Mixed_5a, self).__init__()
        self.conv = BasicConv2d(192, 192, kernel_size=3, stride=2)
        self.maxpool = nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        x0 = self.conv(x)
        x1 = self.maxpool(x)
        out = torch.cat((x0, x1), 1)
        return out


class Inception_A(nn.Module):
    def __init__(self):
        super(Inception_A, self).__init__()
        self.branch0 = BasicConv2d(384, 96, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(384, 64, kernel_size=1, stride=1), BasicConv2d(64, 96, kernel_size=3, stride=1, padding=1)
        )

        self.branch2 = nn.Sequential(
            BasicConv2d(384, 64, kernel_size=1, stride=1),
            BasicConv2d(64, 96, kernel_size=3, stride=1, padding=1),
            BasicConv2d(96, 96, kernel_size=3, stride=1, padding=1),
        )

        self.branch3 = nn.Sequential(
            nn.AvgPool2d(3, stride=1, padding=1, count_include_pad=False),
            BasicConv2d(384, 96, kernel_size=1, stride=1),
        )

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = self.branch3(x)
        out = torch.cat((x0, x1, x2, x3), 1)
        return out


class Reduction_A(nn.Module):
    def __init__(self):
        super(Reduction_A, self).__init__()
        self.branch0 = BasicConv2d(384, 384, kernel_size=3, stride=2)

        self.branch1 = nn.Sequential(
            BasicConv2d(384, 192, kernel_size=1, stride=1),
            BasicConv2d(192, 224, kernel_size=3, stride=1, padding=1),
            BasicConv2d(224, 256, kernel_size=3, stride=2),
        )

        self.branch2 = nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        out = torch.cat((x0, x1, x2), 1)
        return out


class Inception_B(nn.Module):
    def __init__(self):
        super(Inception_B, self).__init__()
        self.branch0 = BasicConv2d(1024, 384, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(1024, 192, kernel_size=1, stride=1),
            BasicConv2d(192, 224, kernel_size=(1, 7), stride=1, padding=(0, 3)),
            BasicConv2d(224, 256, kernel_size=(7, 1), stride=1, padding=(3, 0)),
        )

        self.branch2 = nn.Sequential(
            BasicConv2d(1024, 192, kernel_size=1, stride=1),
            BasicConv2d(192, 192, kernel_size=(7, 1), stride=1, padding=(3, 0)),
            BasicConv2d(192, 224, kernel_size=(1, 7), stride=1, padding=(0, 3)),
            BasicConv2d(224, 224, kernel_size=(7, 1), stride=1, padding=(3, 0)),
            BasicConv2d(224, 256, kernel_size=(1, 7), stride=1, padding=(0, 3)),
        )

        self.branch3 = nn.Sequential(
            nn.AvgPool2d(3, stride=1, padding=1, count_include_pad=False),
            BasicConv2d(1024, 128, kernel_size=1, stride=1),
        )

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = self.branch3(x)
        out = torch.cat((x0, x1, x2, x3), 1)
        return out


class Reduction_B(nn.Module):
    def __init__(self):
        super(Reduction_B, self).__init__()

        self.branch0 = nn.Sequential(
            BasicConv2d(1024, 192, kernel_size=1, stride=1), BasicConv2d(192, 192, kernel_size=3, stride=2)
        )

        self.branch1 = nn.Sequential(
            BasicConv2d(1024, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 256, kernel_size=(1, 7), stride=1, padding=(0, 3)),
            BasicConv2d(256, 320, kernel_size=(7, 1), stride=1, padding=(3, 0)),
            BasicConv2d(320, 320, kernel_size=3, stride=2),
        )

        self.branch2 = nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        out = torch.cat((x0, x1, x2), 1)
        return out


class Inception_C(nn.Module):
    def __init__(self):
        super(Inception_C, self).__init__()

        self.branch0 = BasicConv2d(1536, 256, kernel_size=1, stride=1)

        self.branch1_0 = BasicConv2d(1536, 384, kernel_size=1, stride=1)
        self.branch1_1a = BasicConv2d(384, 256, kernel_size=(1, 3), stride=1, padding=(0, 1))
        self.branch1_1b = BasicConv2d(384, 256, kernel_size=(3, 1), stride=1, padding=(1, 0))

        self.branch2_0 = BasicConv2d(1536, 384, kernel_size=1, stride=1)
        self.branch2_1 = BasicConv2d(384, 448, kernel_size=(3, 1), stride=1, padding=(1, 0))
        self.branch2_2 = BasicConv2d(448, 512, kernel_size=(1, 3), stride=1, padding=(0, 1))
        self.branch2_3a = BasicConv2d(512, 256, kernel_size=(1, 3), stride=1, padding=(0, 1))
        self.branch2_3b = BasicConv2d(512, 256, kernel_size=(3, 1), stride=1, padding=(1, 0))

        self.branch3 = nn.Sequential(
            nn.AvgPool2d(3, stride=1, padding=1, count_include_pad=False),
            BasicConv2d(1536, 256, kernel_size=1, stride=1),
        )

    def forward(self, x):
        x0 = self.branch0(x)

        x1_0 = self.branch1_0(x)
        x1_1a = self.branch1_1a(x1_0)
        x1_1b = self.branch1_1b(x1_0)
        x1 = torch.cat((x1_1a, x1_1b), 1)

        x2_0 = self.branch2_0(x)
        x2_1 = self.branch2_1(x2_0)
        x2_2 = self.branch2_2(x2_1)
        x2_3a = self.branch2_3a(x2_2)
        x2_3b = self.branch2_3b(x2_2)
        x2 = torch.cat((x2_3a, x2_3b), 1)

        x3 = self.branch3(x)

        out = torch.cat((x0, x1, x2, x3), 1)
        return out


class InceptionV4(nn.Module):
    def __init__(self, num_classes=1001):
        super(InceptionV4, self).__init__()
        # Special attributs
        self.input_space = None
        self.input_size = (299, 299, 3)
        self.mean = None
        self.std = None
        # Modules
        self.features = nn.Sequential(
            BasicConv2d(3, 32, kernel_size=3, stride=2),  # 0, layer0
            BasicConv2d(32, 32, kernel_size=3, stride=1),  # 1, layer0
            BasicConv2d(32, 64, kernel_size=3, stride=1, padding=1),  # 2
            Mixed_3a(),  # 3
            Mixed_4a(),  # 4
            Mixed_5a(),  # 5
            Inception_A(),  # 6
            Inception_A(),  # 7
            Inception_A(),  # 8
            Inception_A(),  # 9
            Reduction_A(),  # 10 # Mixed_6a
            Inception_B(),  # 11
            Inception_B(),  # 12
            Inception_B(),  # 13
            Inception_B(),  # 14
            Inception_B(),  # 15
            Inception_B(),  # 16
            Inception_B(),  # 17
            Reduction_B(),  # 18  # Mixed_7a
            Inception_C(),  # 19
            Inception_C(),  # 20
            Inception_C(),  # 21
        )
        self.last_linear = nn.Linear(1536, num_classes)

    def logits(self, features):
        # Allows image of any size to be processed
        adaptiveAvgPoolWidth = features.shape[2]
        x = F.avg_pool2d(features, kernel_size=adaptiveAvgPoolWidth)
        x = x.view(x.size(0), -1)
        x = self.last_linear(x)
        return x

    def forward(self, x):
        x = self.features(x)
        x = self.logits(x)
        return x


def inceptionv4(num_classes=1000, pretrained="imagenet"):
    if pretrained:
        settings = pretrained_settings["inceptionv4"][pretrained]
        assert num_classes == settings["num_classes"], "num_classes should be {}, but is {}".format(
            settings["num_classes"], num_classes
        )

        # both 'imagenet'&'imagenet+background' are loaded from same parameters
        model = InceptionV4(num_classes=1001)
        model.load_state_dict(model_zoo.load_url(settings["url"]))

        if pretrained == "imagenet":
            new_last_linear = nn.Linear(1536, 1000)
            new_last_linear.weight.data = model.last_linear.weight.data[1:]
            new_last_linear.bias.data = model.last_linear.bias.data[1:]
            model.last_linear = new_last_linear

        model.input_space = settings["input_space"]
        model.input_size = settings["input_size"]
        model.input_range = settings["input_range"]
        model.mean = settings["mean"]
        model.std = settings["std"]
    else:
        model = InceptionV4(num_classes=num_classes)
    return model



import torch.nn.init
from torch import nn

__all__ = ["conv1x1", "conv3x3"]


def conv1x1(in_channels: int, out_channels: int, groups=1, bias=True) -> nn.Conv2d:
    conv = nn.Conv2d(in_channels, out_channels, kernel_size=1, groups=groups, bias=bias)
    if bias:
        torch.nn.init.zeros_(conv.bias)
    return conv


def conv3x3(in_channels: int, out_channels: int, stride=1, groups=1, bias=True) -> nn.Conv2d:
    conv = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, groups=groups, bias=bias)
    if bias:
        torch.nn.init.zeros_(conv.bias)
    return conv



from .common import GenericTimmEncoder, make_n_channel_input_std_conv

__all__ = [
    "NFNetF0Encoder",
    "NFNetF1Encoder",
    "NFNetF2Encoder",
    "NFNetF3Encoder",
    "NFNetF4Encoder",
    "NFNetF5Encoder",
    "NFNetF6Encoder",
    "NFNetF7Encoder",
]


class NFNetF0Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f0(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF1Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f1(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF2Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f2(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF3Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f3(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF4Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f4(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF5Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f5(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF6Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f6(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self


class NFNetF7Encoder(GenericTimmEncoder):
    def __init__(self, pretrained=True, layers=None):
        from timm.models import nfnet

        encoder = nfnet.nfnet_f7(pretrained=pretrained, features_only=True)
        super().__init__(encoder, layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.stem_conv1 = make_n_channel_input_std_conv(
            self.encoder.stem_conv1, input_channels, mode, **kwargs
        )
        return self



import numpy as np


def naive_bucket_assignment(costs: np.ndarray, num_buckets: int) -> np.ndarray:
    ordered_indexes = np.argsort(costs)
    assignment = ordered_indexes % num_buckets
    return assignment


def compute_bucket_imbalance_score(costs, assignment):
    buckets = np.unique(assignment)
    costs_per_bucket = []
    for bucket in buckets:
        costs_per_bucket.append(np.sum(costs[assignment == bucket]))
    return np.std(costs_per_bucket)


def random_bucket_assignment(costs: np.ndarray, num_buckets: int, max_iterations: int) -> np.ndarray:
    best_indexes = naive_bucket_assignment(costs, num_buckets)
    best_cost = compute_bucket_imbalance_score(costs, best_indexes)

    for _ in range(max_iterations):
        new_indexes = np.random.permutation(best_indexes)
        new_cost = compute_bucket_imbalance_score(costs, new_indexes)
        if new_cost < best_cost:
            best_indexes = new_indexes
            best_cost = new_cost

    return best_indexes


def filler_bucket_assignment(costs: np.ndarray, num_buckets: int) -> np.ndarray:
    order = np.argsort(-costs)
    current_buckets_cost = np.zeros(num_buckets)
    assignment = np.zeros_like(costs, dtype=int)
    for element_index in order:
        bucket_with_min_cost = np.argmin(current_buckets_cost)
        assignment[element_index] = bucket_with_min_cost
        current_buckets_cost[bucket_with_min_cost] += costs[element_index]

    return assignment



from torch import nn, Tensor

from pytorch_toolbelt.modules.normalization import instantiate_normalization_block, NORM_BATCH
from pytorch_toolbelt.modules.activations import instantiate_activation_block

__all__ = ["DepthwiseSeparableConv2d", "DepthwiseSeparableConv2dBlock"]


class DepthwiseSeparableConv2d(nn.Module):
    """
    Depth-wise separable convolution.
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, padding=1, dilation=1, groups=1, bias=True):
        super(DepthwiseSeparableConv2d, self).__init__()
        self.depthwise = nn.Conv2d(
            in_channels,
            in_channels,
            kernel_size=kernel_size,
            dilation=dilation,
            padding=padding,
            stride=stride,
            bias=bias,
            groups=in_channels,
        )
        self.pointwise = nn.Conv2d(in_channels, out_channels, kernel_size=1, groups=groups, bias=bias)

    def forward(self, x):
        out = self.depthwise(x)
        out = self.pointwise(out)
        return out


class DepthwiseSeparableConv2dBlock(nn.Module):
    """
    Depthwise seperable convolution with batchnorm and activation.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        activation: str,
        kernel_size: int = 3,
        stride=1,
        padding=1,
        dilation=1,
        normalization: str = NORM_BATCH,
    ):
        super(DepthwiseSeparableConv2dBlock, self).__init__()
        self.depthwise = DepthwiseSeparableConv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=False,
        )
        self.bn = instantiate_normalization_block(normalization, out_channels)
        self.act = instantiate_activation_block(activation, inplace=True)

    def forward(self, x: Tensor) -> Tensor:
        x = self.depthwise(x)
        x = self.bn(x)
        return self.act(x)
